<?php

declare(strict_types=1);

namespace Drupal\decoupled_blocks\Controller;

use Drupal\block\BlockRepositoryInterface;
use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Condition\ConditionAccessResolverTrait;
use Drupal\Core\Condition\ConditionPluginCollection;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Path\PathMatcherInterface;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\Core\Url;
use Drupal\node\NodeInterface;
use Drupal\path_alias\AliasManagerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\HttpKernel\HttpKernelInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Returns responses for decoupled_blocks routes.
 */
class DecoupledBlocksController extends ControllerBase {

  use ConditionAccessResolverTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The theme manager.
   *
   * @var \Drupal\Core\Theme\ThemeManagerInterface
   */
  protected $themeManager;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $request_stack;

  /**
   * The HTTP kernel.
   *
   * @var \Symfony\Component\HttpKernel\HttpKernelInterface
   */
  protected $httpKernel;

  /**
   * The path alias manager.
   *
   * @var \Drupal\path_alias\AliasManagerInterface
   */
  protected $pathAliasManager;

  /**
   * The path matcher.
   *
   * @var \Drupal\Core\Path\PathMatcherInterface
   */
  protected $pathMatcher;

  /**
   * The plugin manager condition.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $pluginManagerCondition;

  /**
   * {@inheritdoc}
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, ThemeManagerInterface $themeManager, RequestStack $request_stack,
    HttpKernelInterface $httpKernel, AliasManagerInterface $aliasManager, PathMatcherInterface $pathMatcher, PluginManagerInterface $pluginManagerCondition
  ) {
    $this->entityTypeManager = $entityTypeManager;
    $this->themeManager = $themeManager;
    $this->request_stack = $request_stack;
    $this->httpKernel = $httpKernel;
    $this->pathAliasManager = $aliasManager;
    $this->pathMatcher = $pathMatcher;
    $this->pluginManagerCondition = $pluginManagerCondition;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('theme.manager'),
      $container->get('request_stack'),
      $container->get('http_kernel.basic'),
      $container->get('path_alias.manager'),
      $container->get('path.matcher'),
      $container->get('plugin.manager.condition'),
    );
  }

  /**
   * Builds the blocks response for current page.
   */
  public function index(Request $request): JsonResponse {
    $results = [];
    $blocks = $this->getBlocks();

    $path = $request->query->get('path');
    if (empty($path)) {
      throw new NotFoundHttpException("Page not found");
    }

    $path = $path === '/' ? $path : rtrim($path, '/');

    if (!empty($blocks)) {
      $blocks = array_filter($blocks, function($block) {
        return (strpos($block->getPluginId(), 'block_content:') !== FALSE);
      });
      foreach ($blocks as $id => $block) {
        $isVisible = $this->isBlockVisibleForGivenPath($block, $path);
        if ($isVisible) {
          $uuid = $block->uuid();
          $region = $block->getRegion();
          $weight = $block->getWeight();
          $label = $block->label();
          $plugin = $block->getPluginId();
          $results[$id] = [
            'region' => $region,
            'label' => $label,
            'uuid' => $uuid,
            'weight' => $weight,
            'plugin' => $plugin,
            'data' => $this->getContent($block->getPluginId()),
          ];
        }
      }
    }

    return new JsonResponse($results, 200);
  }

  /**
   * Get Content.
   */
  public function getContent(string $pluginId): array {
    [$plugin_type, $plugin_uuid] = explode(':', $pluginId);
    $data = [];
    if ($plugin_type === 'block_content') {
      $data = $this->getContentBlock($plugin_uuid);
    }
    return $data;
  }

  /**
   * Get block content.
   */
  public function getContentBlock(string $plugin_uuid): array {
    $contentBlock = $this->entityTypeManager
      ->getStorage('block_content')
      ->loadByProperties(['uuid' => $plugin_uuid]);
    $json_api_path_prefix = 'jsonapi';
    if (!empty($contentBlock)) {
      if (is_array($contentBlock)) {
        $contentBlock = reset($contentBlock);
        $uuid = $contentBlock->uuid();
        $paragraph_field_name = "field_paragraphs";
        $this->moduleHandler()
          ->alter('decoupled_blocks_paragraph_field_name', $paragraph_field_name);
        $paragraphs = $contentBlock->get()
          ->referencedEntities();
        $paragraph = reset($paragraphs);
        $paragraphType = $paragraph->getType();
        $filters['fields'] = [
          "block_content--template_block" => "block_machine_name",
        ];
        $this->moduleHandler()
          ->alter('decoupled_blocks_filters', $paragraphType, $filters);
        $endpoint = sprintf('/%s/block_content/template_block/%s/%s', $json_api_path_prefix, $uuid, "field_paragraphs");

        $request = Request::create(
          $endpoint,
          'GET',
          $filters,
          $this->request_stack->getCurrentRequest()->cookies->all(),
          [],
          $this->request_stack->getCurrentRequest()->server->all()
        );

        $response = $this->httpKernel->handle($request, HttpKernelInterface::SUB_REQUEST);

        return ["data" => json_decode($response->getContent(), TRUE)];
      }
    }
    return [];
  }

  /**
   * Get Block of the current active theme.
   */
  protected function getBlocks() {
    $theme = $this->themeManager->getActiveTheme()->getName();
    $regions = system_region_list($theme, BlockRepositoryInterface::REGIONS_VISIBLE);

    $blocks = [];
    foreach ($regions as $key => $region) {
      $region_blocks = $this->entityTypeManager->getStorage('block')
        ->loadByProperties([
          'theme' => $theme,
          'region' => $key,
          'status' => 1,
        ]);

      if (!empty($region_blocks)) {
        $region_blocks = (array) $region_blocks;
        $blocks = array_merge($blocks, $region_blocks);
      }
    }

    return $blocks;
  }

  /**
   * Visibility check.
   */
  protected function isBlockVisibleForGivenPath($block, $path) {
    $visibility = $block->getVisibility();
    if (empty($visibility)) {
      return TRUE;
    }

    $path_visibility = TRUE;
    $request_path = $visibility['request_path'] ?? [];
    if (!empty($request_path)) {
      $negate = !empty($request_path['negate']);
      if (empty($request_path['pages'])) {
        $path_visibility = !$negate;
      }
      else {
        $path_alias = mb_strtolower($this->pathAliasManager->getAliasByPath($path));
        $path_alias = $path_alias === "/" ? "<front>" : $path_alias;
        $match = $this->pathMatcher->matchPath($path_alias, $request_path['pages']) || (($path != $path_alias)
            && $this->pathMatcher->matchPath($path, $request_path['pages']));
        $path_visibility = $negate ? !$match : $match;
      }
      unset($visibility['request_path']);
    }

    $visibility_collection = new ConditionPluginCollection($this->pluginManagerCondition, $visibility);
    $conditions = [];

    foreach ($visibility_collection as $id => $condition) {
      $conditions[$id] = $condition;
      if (in_array($id, ['entity_bundle:node', 'node_type'])) {
        $node = $this->getNodeFromPath($path);
        if ($node instanceof NodeInterface) {
          $condition->setContextValue('node', $node);
        }
      }

      if ($id === 'user_role') {
        $condition->setContextValue('user', $this->currentUser()->getAccount());
      }
    }
    return $this->resolveConditions($conditions, 'and') && $path_visibility;
  }

  /**
   * Get node from path.
   */
  private function getNodeFromPath($path) {
    $node = &drupal_static(__FUNCTION__, NULL);
    if ($node instanceof NodeInterface) {
      return $node;
    }
    try {
      $params = Url::fromUri("internal:" . $path)->getRouteParameters();
      $node_id = $params['node'] ?? NULL;
      if (isset($node_id)) {
        $node = $this->entityTypeManager->getStorage('node')->load($node_id);
        if ($node instanceof NodeInterface) {
          $current_lang = $this->languageManager()
            ->getCurrentLanguage()
            ->getId();
          if ($node->hasTranslation($current_lang)) {
            $node = $node->getTranslation($current_lang);
          }

          $can_view = $node->access('view', NULL, TRUE);
          if (!$can_view->isAllowed()) {
            return NULL;
          }
        }

        return $node;
      }
    }
    catch (\Exception $e) {
      return NULL;
    }
  }

}
